#ifndef CORE_LOGGER_HPP
#define CORE_LOGGER_HPP

#include "core/utils/ArrayString.hpp"

namespace Core::Logger {
    enum class Level { ERROR, WARNING, INFO, DEBUG };
    extern Level level;

    [[maybe_unused]] static constexpr const char* COLOR_RED = "\33[1;31m";
    [[maybe_unused]] static constexpr const char* COLOR_YELLOW = "\33[1;33m";
    [[maybe_unused]] static constexpr const char* COLOR_GRAY = "\33[1;37m";
    [[maybe_unused]] static constexpr const char* COLOR_GREEN = "\33[1;32m";
    [[maybe_unused]] static constexpr const char* COLOR_RESET = "\33[0m";

    const char* getFileName(const char* path);

    // aborts on critical logging failure
    template<typename... Args>
    void log(Level l, const char* file, int line, const char* prefix,
             const char* tag, const char* format, Args&&... args) {
        if(Core::Logger::level < l) {
            return;
        }
        file = getFileName(file);
        Core::ArrayString<2048> s;
        s.append(prefix).append(tag).append("#:# | ");
        s.format(file, line);
        s.append(format);
        s.format(Core::forward<Args>(args)...);
        s.append(COLOR_RESET);
        s.printLine();
    }

    template<typename... Args>
    void log(const char* prefix, const char* format, Args&&... args) {
        Core::ArrayString<2048> s;
        s.append(prefix).append(format);
        s.format(Core::forward<Args>(args)...);
        s.append(COLOR_RESET).printLine();
    }
}

#if defined(CORE_LOG_LEVEL) && CORE_LOG_LEVEL >= 1
#define CORE_LOG_ERROR(format, ...)                                            \
    log(Core::Logger::Level::ERROR, __FILE__, __LINE__,                        \
        Core::Logger::COLOR_RED, "[ERROR] ",                                   \
        format __VA_OPT__(, ) __VA_ARGS__);
#else
#define CORE_LOG_ERROR(format, ...)
#endif

#if defined(CORE_LOG_LEVEL) && CORE_LOG_LEVEL >= 2
#define CORE_LOG_WARNING(format, ...)                                          \
    log(Core::Logger::Level::WARNING, __FILE__, __LINE__,                      \
        Core::Logger::COLOR_YELLOW, "[WARNING] ",                              \
        format __VA_OPT__(, ) __VA_ARGS__);
#else
#define CORE_LOG_WARNING(format, ...)
#endif

#if defined(CORE_LOG_LEVEL) && CORE_LOG_LEVEL >= 3
#define CORE_LOG_INFO(format, ...)                                             \
    log(Core::Logger::Level::INFO, __FILE__, __LINE__,                         \
        Core::Logger::COLOR_GRAY, "[INFO] ",                                   \
        format __VA_OPT__(, ) __VA_ARGS__);
#else
#define CORE_LOG_INFO(format, ...)
#endif

#if defined(CORE_LOG_LEVEL) && CORE_LOG_LEVEL >= 4
#define CORE_LOG_DEBUG(format, ...)                                            \
    log(Core::Logger::Level::DEBUG, __FILE__, __LINE__,                        \
        Core::Logger::COLOR_GREEN, "[DEBUG] ",                                 \
        format __VA_OPT__(, ) __VA_ARGS__);
#else
#define CORE_LOG_DEBUG(format, ...)
#endif

#endif
